/*-
 * UNPUBLISHED SOURCE CODE
 *
 * Copyright © 2017, Brocade Communications Systems, Incorporated.
 * ALL RIGHTS RESERVED.
 *
 * BROCADE HIGHLY CONFIDENTIAL TRADE SECRET INFORMATION
 *
 * Access or possession of this material grants you no right or license,
 * express, implied, statutory or otherwise, under any Brocade patent,
 * copyright, trade secret right or other intellectual property right.
 * Any such license must be contained in a written license agreement
 * signed by an authorized officer of Brocade.  Additionally, access or
 * possession of this material is limited to Brocade employees with a
 * need to know or named employees of Brocade vendors authorized by
 * Brocade in writing to access this material.
 *
 * Restricted Rights: Use, duplication, or disclosure by the U.S.
 * Government is subject to restrictions of FAR 52.227-19 and its
 * successors, or (c)(1)(ii) of the Rights in Technical Data and
 * Computer Software clause at DFAR 252.227-7013 and its successors.
 *
 * Brocade Communications Systems, Inc.,
 * 110 Holger Way, San Jose, CA 95134
 */
/**
* Author : gneelagi
* Version : 1.0
*/
import path from 'path';
import fs from 'fs';
import env from 'os';
import stylegenerator, { compile as stylecompiler } from '../style/style-generator.js';
import cliArgParser from 'commander';
import colors from 'colors';
import React from 'react';
import ReactDOMServer from 'react-dom/server';
import concatFiles from 'concat-files';
import { TitlePanelFactory } from '../table/title-panel-enhanced.js';
import { FilterPanel } from '../table/filter-panel.js';
import { NotesPanel } from '../table/notes-panel.js';
import { ReportTitle } from '../table/simple-report-title.js';
import { WidgetTitle } from '../table/simple-widget-title.js';
import { KeyValueList } from '../table/simple-keyvalue-pair.js';
import { ValueList } from '../table/simple-valuelist.js';
import deepmerge from 'deepmerge';
import { loggerInstance } from './logger.js';

const logger = (label) => {
  const splitLabel = label.split(/lib./);
  const formattedLabel = splitLabel.length == 2 ? splitLabel[1] : splitLabel;
  return {
      info: function() {
          // convert arguments to real array
          var args = Array.prototype.slice.call(arguments);
          var logMessage = formattedLabel.concat(" - ").concat(args.join(" "));
          loggerInstance.log.apply(loggerInstance, ["info", logMessage]);    
      },
      warn: function() {
          var args = Array.prototype.slice.call(arguments);
          var logMessage = formattedLabel.concat(" - ").concat(args.join(" "));
          loggerInstance.log.apply(loggerInstance, ["warn", logMessage]);
      },
      error: function() {
          var args = Array.prototype.slice.call(arguments);
          var logMessage = formattedLabel.concat(" - ").concat(args.join(" "));
          loggerInstance.log.apply(loggerInstance, ["error", logMessage]);
      },
      debug: function() {
        var args = Array.prototype.slice.call(arguments);
        var logMessage = formattedLabel.concat(" - ").concat(args.join(" "));
        loggerInstance.log.apply(loggerInstance, ["debug", logMessage]);
    }
  }
};

var locale, defaultLimit = {}, chartFontFamily, debug = false;
var timeout = {};
const messageFunctionPattern = new RegExp("^\\{message\\((.*)\\)\\}$"); 
const LOGGER = logger(__filename);
const reportConfig = {};
const reportConfigProvider = ()=> {
  return reportConfig;
}

// ideally we should avoid globals, reportConfig is to have global context.
// FIXME find a way to pass the configuration available at report level
// to individual widget level, by creating a visitor like we have for title & filter

const resolveMessage = (messageReference) => {
  const parsedMessage = messageFunctionPattern.exec(messageReference) || [];
  let message;
  if (parsedMessage.length > 1) {
    const messageId = parsedMessage[1];
    message = locale.messages[messageId];
  } else {
    message = messageReference;
  }

  return message;
}

const readJsonFile = (filePath) => {
  let fileAsBytes = fs.readFileSync(filePath);
  if (fileAsBytes) {
    return JSON.parse(fileAsBytes);
  }
  const empty = "";
  return empty;
}

const readCssFile = (filePath) => {
  let fileAsBytes = fs.readFileSync(filePath);
  return fileAsBytes.toString();
}
const args = {};

const init = () => {
  const platform = env.platform();
  cliArgParser
    .version('1.0.0')
    .option('-d ,--reportdir <arg>', 'Report directory')
    .option('-i ,--debug', 'Inspect performance')
    .option('-c ,--maxchartssupportedperwidget <arg>', 'Max. Charts Supported / Widget')
    .option('-s ,--maxseriessupportedperchart <arg>', 'Max. Series Supported / Chart')
    .option('-r ,--maxrowssupportedpertable <arg>', 'Max. Rows Supported / Table')
    .option('-p ,--propdir <arg>', 'properties.json directory')
    .option('-S ,--maxseriessupportedperwidget <arg>', 'Max. Series Supported / Widget')
    .option('-o ,--port <arg>', 'port to start')
    .option('-e ,--chromeexecutablepath <arg>','chrome browser, to be launched in headless mode, path that ends with chrome.exe for windows')

    .parse(process.argv);
  args.reportdir = cliArgParser.reportdir || '';
  if (platform === 'linux') {
    // in linux  "folder with space" becomes '"folder with space"' 
    // that is the the input with double quote is surrounded by single quote
    // this leads to in-appropriate path.
    args.reportdir = args.reportdir.replace(/[\"]/g, "");
  }
  args.debug = cliArgParser.debug;
  args.maxchartssupportedperwidget = cliArgParser.maxchartssupportedperwidget;
  args.maxseriessupportedperchart = cliArgParser.maxseriessupportedperchart;
  args.maxrowssupportedpertable = cliArgParser.maxrowssupportedpertable;
  args.maxseriessupportedperwidget = cliArgParser.maxseriessupportedperwidget;
  initConfig(cliArgParser.propdir);
  args.debug = args.debug || debug;
  args.chromeexecutablepath = cliArgParser.chromeexecutablepath;
  if (args.maxseriessupportedperwidget) {
    try {
      let maxSeriesSupportedPerWidget = args.maxseriessupportedperwidget || defaultLimit.maxSeriesSupportedPerWidget;
      let maxSeriesSupportedPerChart = args.maxseriessupportedperchart || defaultLimit.maxSeriesSupportedPerChart;
      maxSeriesSupportedPerChart = parseInt(maxSeriesSupportedPerChart);
      maxSeriesSupportedPerWidget = parseInt(maxSeriesSupportedPerWidget);
      args.maxchartssupportedperwidget = parseInt(maxSeriesSupportedPerWidget / maxSeriesSupportedPerChart);
    } catch (e) {
      //TODO set proper level & control log based on level
      LOGGER.info('Error while parsing command line args ',args);
      LOGGER.info(e.stack);
    }
  }

  // stylecompiler();  
  const styleplatform = 'win32'; // let us use the same style across platform
  // else the style tailored for linux server accessed in non-linux machine 
  // will be having different look & feel(due to the DPI difference in windows,linux)
  // stylegenerator.generate('screen', args.reportdir, styleplatform);
  // stylegenerator.generate('A4', args.reportdir, styleplatform);
  return { reportdir: args.reportdir, port: cliArgParser.port };
}

const initConfig = (confFilePath) => {
  let conf = {};
  if (confFilePath) {
    conf = readJsonFile(confFilePath);
    LOGGER.info('time zone ', conf.localTimeZone);
  }
  locale = conf.locale;
  defaultLimit = conf.limits;
  chartFontFamily = conf.chartFontFamily;
  timeout = conf.timeout || {};
  debug = conf.debug || false;
}

const program = () => {
  args.reportdir = '';
  return args;
}

const limits = () => {
  const fromCli = program();
  const limit = {
    "maxChartsSupportedPerWidget": fromCli.maxchartssupportedperwidget
      || defaultLimit.maxChartsSupportedPerWidget,
    "maxSeriesSupportedPerChart": fromCli.maxseriessupportedperchart
      || defaultLimit.maxSeriesSupportedPerChart,
    "maxRowsSupportedPerTable": fromCli.maxrowssupportedpertable
      || defaultLimit.maxRowsSupportedPerTable
  };
  return limit;
}

const timeouts = ()=>{
  return timeout;
}

const messages = () => {
  if (locale) {
    return locale.messages || [];
  }
  return [];
}
const widgetsByType = (schema) => {
  const widgetsByType = { chart: {}, headerDetailsFooter: {} };
  const widgets = schema.template.widgets;
  let headerDetailsFooterWidgets = widgetsByType.headerDetailsFooter;
  let chartWidgets = widgetsByType.chart;
  let widgetByType = (widgetId) => {
    const view = widgets[widgetId].view;
    if (view) {
      if (view.as === 'chart') {
        chartWidgets[widgetId] = widgets[widgetId];
      } else {
        headerDetailsFooterWidgets[widgetId] = widgets[widgetId];
      }
    }
  }
  Object.keys(widgets).map(widgetByType);
  return widgetsByType;
}

const DatasetInfo = () => {
  // for now we are considering only the table ,rows,columns count
  // as the dataset that we build will be of that form
  // each table will have same no.of columns
  // likewise for chart each series will have same points
  const tableInfo = { tables: 0, rows: 0, columns: 0 };
  const chartInfo = { charts: 0, series: 0, points: 0 };
  const format = () => {
    // todo receive input template literal
    return `report has 
  ${datasetInfo.tableInfo.tables} table(s) with ${datasetInfo.tableInfo.rows} rows of ${datasetInfo.tableInfo.columns} columns  
  ${datasetInfo.chartInfo.charts} chart(s) with ${datasetInfo.chartInfo.series} series of ${datasetInfo.chartInfo.points} points    `;
  }
  const datasetInfo = { tableInfo, chartInfo, format };
  return datasetInfo;
}
const TimeInfo = () => {
  const html = 0;
  const report = 0;
  const pdf = 0;
  const format = () => {
    // todo receive input template literal
    const htmlFormatted = colors.green(timeInfo.html.toFixed(2));
    const pdfFormatted = colors.green(timeInfo.pdf.toFixed(2));
    const reportFormatted = colors.green(timeInfo.report.toFixed(2));
    return `html | pdf | report generation time (in seconds)
   ${htmlFormatted} | ${pdfFormatted} | ${reportFormatted} `;
  }
  const timeInfo = { html, pdf, report, format };
  return timeInfo;
}
const MemoryInfo = () => {
  const mb = 1024 * 1024;
  const nodejs = 0;
  const phantomjsChartWorkers = 1;
  const phantomjsChartWorkerMem = 60 * mb;
  const phantomjsChart = 0;
  const phantomjsPdf = 0;
  const format = () => {
    // todo receive input template literal

    const nodejsMem = memoryInfo.nodejs / mb;
    const phantomjsChartMem = memoryInfo.phantomjsChart / mb;
    const phantomjsPdfMem = memoryInfo.phantomjsPdf / mb;

    const nodejsMemFormatted = colors.green(nodejsMem.toFixed(2));
    const phantomjsChartMemFormatted = colors.green(phantomjsChartMem.toFixed(2));
    const phantomjsPdfMemFormatted = colors.green(phantomjsPdfMem.toFixed(2));

    const nodejsProcesses = 1;
    const phantomjsChartWorkerProcesses = 1;
    const phantomjsPdfProcess = 1;
    return `report generator has 
   ${nodejsProcesses} nodejs for html conversion
   ${phantomjsChartWorkerProcesses} phantomjs for chart/svg conversion
   ${phantomjsPdfProcess} phantomjs for html-to-pdf conversion
nodejs | phantomjs/chart | phantomjs/pdf process'
  total memory allocated (in MiB)
   ${nodejsMemFormatted} | ${phantomjsChartMemFormatted} | ${phantomjsPdfMemFormatted}
 
Note:
  phantomjs memory allocation for chart/svg conversion is arrived based
  on calculation as the generator don't have a process handle of this.
`;
  }
  const memoryInfo = { nodejs, phantomjsChart, phantomjsChartWorkerMem, phantomjsPdf, format };
  return memoryInfo;
}

const saveAsFile = (filePath, data) => {
  fs.writeFileSync(filePath, data);
}
const toHtml = (ReactComponent, props) => {
  const html = ReactDOMServer.renderToStaticMarkup(<ReactComponent {...props} />);
  return html;
}
const saveAsHtml = (ReactComponent, props) => {
  const html = toHtml(ReactComponent, props);
  const widgetKey = props.widgetKey;
  const filePath = path.join(props.reportdir, widgetKey + '.html');
  saveAsFile(filePath, html);
  //  const SUCCESS = 0; // it is sync else this might be called in a async
  props.callback(null, filePath); //TODO handle failure case             
}

const saveHeaderAsHtml = (props) => {
  // we should have the title, filter(if any) & notes as part of the header to the component/container.  
  props = props || { widgetKey: 'parent' };
  props.widgetKey = props.widgetKey + ';header';
  props.callback = props.callback || ((err, result) => {
    LOGGER.info(result);
  });
  props.title = props.title || [
  ];
  props.filters = props.filters || []
  props.note = props.note || [{ text: ['Notes goes here'], className: 'table-note' }];

  const title = props.title; // disable headers for now
  const filters = props.filters;
  const note = props.note;
  const Header = (props) => {
    const TitlePanel = TitlePanelFactory(props.title);
    return <section><TitlePanel title={title} /><FilterPanel filters={filters} /><NotesPanel note={note} /></section>

  }
  saveAsHtml(Header, props);
}

const saveHeaderAsHtmlNew = (props) => {
  // we should have the title, filter(if any) & notes as part of the header to the component/container.  
  props = props || { widgetKey: 'parent' };
  props.widgetKey = props.widgetKey + ';header';
  props.callback = props.callback || ((err, result) => {
    LOGGER.info(result);
  });
  props.headerType = props.headerType || '';
  props.keyValueList = props.keyValueList || [];
  props.title = props.title || {};
  props.subTitle = props.subTitle || {};
  props.valueList = props.valueList || [];
  props.note = [];
  props.note = props.note || [{ text: ['Notes goes here'], className: 'table-note' }];

  // const note = props.note;
  const Header = (props) => {
    switch (props.headerType) {
      case 'report': {
        if (props.title.text) {
          return <section className="report-header">
            <ReportTitle title={props.title} />
            <KeyValueList keyValueList={props.keyValueList} />
            <NotesPanel note={props.note} />
          </section>;
        }
      }
      default: {
        if (props.title) {
          return <section>
            <WidgetTitle title={props.title} subTitle={props.subTitle} />
            <ValueList valueList={props.valueList} />
            <NotesPanel note={props.note} />
          </section>;
        }
      }
    }
  }
  saveAsHtml(Header, props);
}


const joinFiles = (filePaths, destPath, callback) => {
  concatFiles(filePaths, destPath, callback);
}

const datasetInfo = DatasetInfo();
const timeInfo = TimeInfo();
const memoryInfo = MemoryInfo();


const workspaceLocation = (name) => {
  const timeMilliseconds = new Date().getTime();
  // todo get it from a config properties.
  const folderName = path.join(homeDir, "tmp", "report-" + name + "-" + timeMilliseconds);
  fs.mkdirSync(folderName);
  return folderName;
}

const outdirbase = (name) => {
  const timeMilliseconds = new Date().getTime();
  const folderName = path.join(homeDir, "tmp", "report-" + name + "-" + timeMilliseconds);
  fs.mkdirSync(folderName);
  return folderName;
}

const cleanupWorkspace = (folder) => {
  const files = fs.readdirSync(folder);
  const removeIntermediateFiles = (filename) => {
    if (filename !== 'report.html') {
      fs.unlinkSync(path.join(folder, filename));
    }
  }
  files.forEach(removeIntermediateFiles);
}
const homeDir = path.join(__dirname, '..', '..');

const concatCssFiles = (props) => {
  const conf = { themeDir: path.join(homeDir, 'conf', 'theme'), 'theme': 'default' };
  const defaultFilesToInclude = ['report/report.css', 'table/table.css', 'widget/widget.css', 'title/title.css', 'filter/filter.css'];
  const makeAbsolutePath = (fileName) => {
    return path.join(conf.themeDir, conf.theme, fileName);
  }
  const filesToIncludeFullPath = defaultFilesToInclude.map(makeAbsolutePath);
  const destFileAbsolutePath = makeAbsolutePath('style.css');
  const filesToInclude = props.files || filesToIncludeFullPath;
  const callback = (err) => {
    if (err) {
      //TODO set proper level & control log based on level
      LOGGER.info(err.stack)
    } else {
      let css = fs.readFileSync(destFileAbsolutePath);
      css = '<style>' + css + '</style>';
      fs.writeFileSync(destFileAbsolutePath, css);
    }
  }
  joinFiles(filesToInclude, destFileAbsolutePath, callback);
}

const makeHtmlDocument = (props) => {  
  const css = fs.readFileSync(props.stylecss);
  const html = fs.readFileSync(props.reporthtml);
  const htmlDoc = '<html>' + css + html + '</html>';
  const htmlFile = props.reporthtml;
  fs.writeFileSync(htmlFile, htmlDoc);
}

const mergeJsonDocument = (packagePath, parent, child) => {
  const parentFile = path.join(packagePath, parent);
  const childFile = path.join(packagePath, child);
  const parentJson = readJsonFile(parentFile);
  const childJson = readJsonFile(childFile);

  return Object.assign(parentJson, childJson);

}

const createJsonDocument = (packagePath, childJson) => {
  const parentRef = childJson['__extends__'];
  if (parentRef) {
    delete childJson['__extends__'];
    const parentFile = path.join(packagePath, parentRef + '.json');
    const parentJson = readJsonFile(parentFile);
    const childJsonNew = deepmerge(parentJson, childJson);
    return createJsonDocument(packagePath, childJsonNew);
  } else {
    return childJson;
  }
}

const deepMergeJson = (dest, source) => {
  return deepmerge(dest, source);
}

const styleJson = (packagePath, styleName) => {
  const jsonFile = path.join(packagePath, styleName + '.json');
  const json = readJsonFile(jsonFile);
  return createJsonDocument(packagePath, json);
}
const wrapHtml = (str, htmlFile) => {
  const html = fs.readFileSync(htmlFile);
  const htmlWrapped = str[0] + html + str[1];
  fs.writeFileSync(htmlFile, htmlWrapped);
  return htmlFile;
}

//For rendering html wrto given single/composite widget model holder configuration style
const wrapHtmlWithStyle = (str, style, htmlFile) => {  
  const html = fs.readFileSync(htmlFile);
  const htmlWrapped = str[0] + style + str[1] + html + str[2];
  fs.writeFileSync(htmlFile, htmlWrapped);
  return htmlFile;
}

const copyReportOutputToLocation = (reportGeneratedLocation, reportLocation) => {
  const filesToCopy = ['report.html', 'report.pdf'];
  // const filesToCopy = ['report.html'];
  const copyTo = (sourceFileName) => {
    const sourcePath = path.join(reportGeneratedLocation, sourceFileName);
    const destPath = path.join(reportLocation, sourceFileName);
    const sourceContent = fs.readFileSync(sourcePath);
    fs.writeFileSync(destPath, sourceContent);
  }
  filesToCopy.forEach(copyTo);
}



export { copyReportOutputToLocation }
export { homeDir }
export { wrapHtml }
export { styleJson }
export { deepMergeJson }
export { createJsonDocument }
export { mergeJsonDocument }
export { makeHtmlDocument }
export { concatCssFiles }
export { workspaceLocation }
export { cleanupWorkspace }
export { saveHeaderAsHtml }
export { saveHeaderAsHtmlNew }
export { joinFiles }
export { saveAsHtml }
export { toHtml }
export { saveAsFile }
export { resolveMessage }
export { messages }
export { limits }
export { memoryInfo }
export { timeInfo }
export { datasetInfo }
export { widgetsByType }
export { readJsonFile }
export { readCssFile }
export { init }
export { program }
export { chartFontFamily }
export { wrapHtmlWithStyle}
export { timeouts }
export { reportConfigProvider }
export { logger }

if (require.main == module) {
  // saveHeaderAsHtml();
  // const loc = workspaceLocation("Health Summary");
  // console.log('location ',loc);
  // const folder = "F:/home/govindan/report/report-Health Summary-1510830522097";
  // cleanupWorkspace(folder);
  //  const props = {};
  //  concatCssFiles(props);
  // makeHtmlDocument(props);

  const packagePath = 'F:/report-infra/local/server_new/framework/reportgenerator/conf/theme/default/chart';
  // const lineChartJson = readJsonFile(path.join(packagePath,'line.json'));
  const jsonStyleName = 'dual-line-series';

  const finalJson = styleJson(packagePath, jsonStyleName);

  console.log(finalJson);

  //const finalJson = createJsonDocument(packagePath,lineChartJson);
  //console.log('final json ', finalJson);
}